/*	$Id: qqparse.y,v 1.6 2009/07/30 00:10:05 nis Exp $	*/

%{
/*-
 * Copyright (c) 2005 Shingo Nishioka.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#if ! defined lint
static char rcsid[] = "$Id: qqparse.y,v 1.6 2009/07/30 00:10:05 nis Exp $";
#endif

/* #include <config.h> */

#include <sys/types.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "qcpp.h"

#include <gifnoc.h>

int yyparse(void);

static int yyerror(char *);

static struct attribute *arev(struct attribute *);
static union content **crev(union content **);
static struct element *alloc_element(char *, struct attribute *, union content **);
static struct misc *alloc_comment(char *);
static struct chardata *alloc_chardata(char *);
static struct cexpr *alloc_cexpr(int, void *, int);
static struct cexpr *nconc(struct cexpr *, struct cexpr *);

static struct cexpr *cexpr;

%}

%union {
	char *name;
	struct element *element;
	struct attribute *attributes, attribute;
	union content **contents, *content;
	struct cexpr *cexpr;
}

%start cc

%token <name> ID STRING CTOKEN INUM SPACE COMMENT
%type <name> cexpr_token
%type <cexpr> cc cprog cprog_item cexpr cexpr_item a a_item unquote unquote_splicing unq
%type <element> elem qquote comment
%type <attributes> attrs
%type <attribute> attr
%type <contents> conts
%type <content> cont

%%

cc		:	cprog a {
				$$ = nconc($1, $2);
				cexpr = $$;
			}
		;

cprog		:	/* empty */ {
				$$ = NULL;
			}
		|	cprog a cprog_item {
				$$ = nconc($1, nconc($2, $3));
			}
		;

cprog_item	:	cexpr_item {
				$$ = $1;
			}
		|	';' {
				$$ = alloc_cexpr(CCtoken, strdup(";"), yytokline());
			}
		|	'{' {
				$$ = alloc_cexpr(CCtoken, strdup("{"), yytokline());
			}
		|	'}' {
				$$ = alloc_cexpr(CCtoken, strdup("}"), yytokline());
			}
		;

cexpr		:	/* empty */ {
				$$ = NULL;
			}
		|	cexpr_item a cexpr {
				$$ = nconc($1, nconc($2, $3));
			}
		;

cexpr_item	:	qquote {
				$$ = alloc_cexpr(CQquote, $1, 0);
			}
		|	'#' '#' {
				struct cexpr *t, *u;
				t = alloc_cexpr(CCtoken, strdup("#"), yytokline());
				u = alloc_cexpr(CCtoken, strdup("#"), yytokline());
				t = nconc(u, t);
				$$ = t;
			}
		|	'#' a cexpr_token {
				struct cexpr *t, *u;
				t = alloc_cexpr(CCtoken, $3, yytokline());
				t = nconc($2, t);
				u = alloc_cexpr(CCtoken, strdup("#"), yytokline());
				t = nconc(u, t);
				$$ = t;
			}
		|	'#' a INUM a STRING {
				struct cexpr *t, *u;
				t = alloc_cexpr(CCtoken, $5, yytokline());
				t = nconc($4, t);
				u = alloc_cexpr(CCtoken, $3, yytokline());
				t = nconc(u, t);
				t = nconc($2, t);
				u = alloc_cexpr(CCtoken, strdup("#"), yytokline());
				t = nconc(u, t);
				$$ = t;
				yysetline(atoi($3) - 1);
			}
		|	INUM {
				$$ = alloc_cexpr(CCtoken, $1, yytokline());
			}
		|	cexpr_token {
				$$ = alloc_cexpr(CCtoken, $1, yytokline());
			}
		;

cexpr_token	:	ID {
				$$ = $1;
			}
		|	STRING {
				$$ = $1;
			}
		|	CTOKEN {
				$$ = $1;
			}
		|	'(' {
				$$ = strdup("(");
			}
		|	')' {
				$$ = strdup(")");
			}
		|	',' {
				$$ = strdup(",");
			}
		|	'=' {
				$$ = strdup("=");
			}
		;

qquote  	:  	'`' b elem {
				$$ = $3;
			}
		|	'`' b comment {
				$$ = $3;
			}
		;

elem		:	ID b '(' b attrs ')' b '{' b conts '}' {
				$$ = alloc_element($1, arev($5), crev($10));
			}
		;

comment		:	COMMENT {
				$$ = (struct element *)alloc_comment($1);
			}
		;

attrs		:	/* empty */ {
				struct attribute *as;
				int i;
				as = calloc(128, sizeof *as);	/* XXX */
				for (i = 0; i < 128; i++) as[i].type = 0;
				$$ = as;
			}
		|	attr b attrs {
				struct attribute *a;
				for (a = $3; a->type != 0; a++) ;
				*a = $1;
				$$ = $3;
			}
		;

attr		:	ID b '=' b STRING b attr_sep {
				$$.type = AAttribute;
				$$.name = $1;
				$$.value.string = $5;
			}
		|	ID b '=' b unquote {
				$$.type = AUnquoteVal;
				$$.name = $1;
				$$.value.cexpr = $5;
			}
		|	unquote {
				$$.type = AUnquote;
				$$.name = NULL;
				$$.value.cexpr = $1;
			}
		|	unquote_splicing {
				$$.type = AUnquoteSplicing;
				$$.name = NULL;
				$$.value.cexpr = $1;
			}
		;

conts 		:	/* empty */ {
				union content **cs;
				int i;
				cs = calloc(128, sizeof *cs);	/* XXX */
				for (i = 0; i < 128; i++) cs[i] = NULL;
				$$ = cs;
			}
		|	cont b conts {
				union content **c;
				for (c = $3; *c; c++) ;
				*c = $1;
				$$ = $3;
			}
		;

cont 		:	elem b cont_sep {
				$$ = (union content *)$1;
			}
		|	comment b cont_sep {
				$$ = (union content *)$1;
			}
		|	STRING b cont_sep {
				$$ = (union content *)alloc_chardata($1);
			}
		|	unquote {
				$$ = (union content *)$1;
			}
		|	unquote_splicing {
				$$ = (union content *)$1;
			}
		;

attr_sep	:	/* empty */
		|	';'
		;

cont_sep	:	/* empty */
		|	';'
		;

unquote		:	',' a unq a ';' {
				$3->type = CUnquote;
				$$ = $3;
			}
		;

unquote_splicing:	',' '@' a unq a ';' {
				$4->type = CUnquoteSplicing;
				$$ = $4;
			}
		;

unq		:	cexpr {
				struct cexpr *c;
				c = alloc_cexpr(0, NULL, yytokline());
				$$ = nconc(c, $1);
			}

a		:	/* empty */ {
				$$ = NULL;
			}
		|	a a_item {
				$$ = nconc($1, $2);
			}
		;

a_item		:	SPACE {
				$$ = alloc_cexpr(CCtoken, $1, yytokline());
			}
		|	COMMENT {
				$$ = alloc_cexpr(CCtoken, $1, yytokline());
			}
		;

b		:	/* empty */
		|	b SPACE
		;

%%

static int
yyerror(s)
char *s;
{
	fprintf(stderr, "%s: line %d, column %d, character '%c'\n", s, 
		yyline(),
		yycolumn(),
		yylastchar());
	return 0;
}

struct cexpr *
rqquote(stream)
FILE *stream;
{
	int e;

	cexpr = NULL;
	init_mygetc(stream);
	e = yyparse();
	if (e == 0) {
		return cexpr;
	}
	return NULL;
}

static struct attribute *
arev(a)
struct attribute *a;
{
	struct attribute *p, *q;
	for (q = a; q->type != 0; q++) ;
	for (p = a, q--; p < q; p++, q--) {
		struct attribute t;
		t = *p, *p = *q, *q = t;
	}
	return a;
}

static union content **
crev(c)
union content **c;
{
	union content **p, **q;
	for (q = c; *q; q++) ;
	for (p = c, q--; p < q; p++, q--) {
		union content *t;
		t = *p, *p = *q, *q = t;
	}
	return c;
}

static struct element *
alloc_element(name, as, cs)
char *name;
struct attribute *as;
union content **cs;
{
	union content *c;
	struct element *e;
	c = calloc(1, sizeof *c);
	e = &c->element;
	e->type = CElem;
	e->name = name;
	e->attributes = as;
	e->contents = cs;
	return e;
}

struct chardata *
alloc_chardata(ss)
char *ss;
{
	union content *c;
	struct chardata *s;
	c = calloc(1, sizeof *c);
	s = &c->chardata;
	s->type = CString;
	s->value = ss;
	return s;
}

static struct misc *
alloc_comment(s)
char *s;
{
	union content *c;
	struct misc *m;

	c = calloc(1, sizeof *c);
	m = &c->misc;
	m->type = CComment;
	m->string = s;
	return m;
}

static struct cexpr *
alloc_cexpr(type, ptr, line)
void *ptr;
{
	union content *c;
	struct cexpr *p;
	c = calloc(1, sizeof *c);
	p = &c->cexpr;
	p->type = type;
	switch (type) {
	case 0:
		break;
	case CCtoken:
		p->u.token = ptr;
		break;
	case CQquote:
		p->u.qquote = ptr;
		break;
	default:
		fprintf(stderr, "internal error");
		exit(1);
	}
	p->line = line;
	p->next = NULL;
	return p;
}

static struct cexpr *
nconc(p, q)
struct cexpr *p, *q;
{
	struct cexpr *r;
	if (!p) {
		return q;
	}
	for (r = p; r->next; r = r->next) ;
	r->next = q;
	return p;
}
